/*
 * Automatic channel selector for ieee802.11 A, B and G modes.
 *
 * Copyright (c) 2005 by Wilibox (www.wilibox.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <string.h>
#include <iwlib.h>

#include "logger.h"
#include "selector.h"

/* Quality delta is a max. difference in quality levels between
 * the two channels to consider them as "similar quality" channels. */
static const long QUALITY_DELTA = 10;

/* Wireless interface name */
static char iface[IFNAMSIZ] = {0};

/* IEEE 802.11 mode: A, B or G */
enum ieee_mode_e { IEEE_MODE_NONE, IEEE_MODE_A, IEEE_MODE_B, IEEE_MODE_G };
static int ieee_mode = IEEE_MODE_NONE;

/* Interface states */
enum iface_state_e { IFACE_DOWN = 0, IFACE_UP = 1 };

/* Distances between the channels for various IEEE modes */
static const int CHANNELS_DISTANCE_A = 4;
static const int CHANNELS_DISTANCE_BG = 1;

/* Quality data array */
static channel_quality qual_data[QUAL_DATA_SIZE];

/* Various card related info, including supported channels */
static struct iw_range range;


#if 0 /* NOT USED */

static int card_ch_count = 0; /* count of supported by card channels */
static int card_channels[QUAL_DATA_SIZE]; /* channels supported by card */

/* Hardcoded channels should not be used since every card can have
 * custom channel set. It also depends on a regulatory domain. Wireless
 * extensions should be used to gather a supported channel set. */
static unsigned int supported_channels[] =
{
 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,  /* 2.412 GHz to 2.?? GHz 802.11b band */
 11, 12, 13, 14,

 36, 40, 44, 48, 52, 56, 60, 64, /* 5.15 GHz to 5.35 GHz 802.11a band */
                                 /* Band is common between Europe and the US.
				  * It's used in almost every European
				  * country.*/

 100, 104, 108, 112, 116, 120,   /* 5.47 GHz to 5.725 GHZ 802.11b band */
 124, 128, 132, 136, 140,        /* Band is available in some European countries
                                  * now and is expected to be available in all
                                  * European countries by late 2004. Made
                                  * available in the U.S. in late 2003. */

 149, 153, 157, 161, 165         /* 5.725 GHz to 5.85 GHz 802.11b band */
                                 /* Band is available in U.S., Canada, and China
                                  * but is not permitted in the EU. */
};

#endif



/* Cleanup */
static void init(void)
{
    bzero(qual_data, sizeof(qual_data));
    /* bzero(card_channels, sizeof(card_channels)); */
    bzero(&range, sizeof(struct iw_range));
}

/* Do ioctl command.
 * Returns 1 on success, 0 on error */
static int do_ioctl(const unsigned long cmd, struct iwreq * data)
{
    int skfd, rv;

    skfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (skfd < 0)
    {
        error_errno("do_ioctl: cannot get free socket\n");
        return 0;
    }

    rv = ioctl(skfd, cmd, data);
    close(skfd);
    if (rv < 0)
    {
        error_errno("do_ioctl: send ioctl failed (cmd=0x%04lX, err=%d)\n",
                    cmd, rv);
        return 0;
    }

    return 1;
}

/* Get current card mode: master/managed/etc.
 * Returns 1 on success, 0 on error. */
static int get_mode(long *mode)
{
    struct iwreq wrq;
    int rv;

    memset(&wrq, 0, sizeof(wrq));
    strncpy(wrq.ifr_name, iface, IFNAMSIZ);
    wrq.u.mode = 0;
    rv = do_ioctl(SIOCGIWMODE, &wrq);
    *mode = wrq.u.mode;

    if (!rv)
        return 0;

    return 1;
}

/* Set current card mode: master/managed/etc.
 * Returns 1 on success, 0 on error. */
static int set_mode(const long new_mode)
{
    struct iwreq wrq;
    int rv;

    memset(&wrq, 0, sizeof(wrq));
    strncpy(wrq.ifr_name, iface, IFNAMSIZ);
    wrq.u.mode = new_mode;
    rv = do_ioctl(SIOCSIWMODE, &wrq);

    if (!rv)
        return 0;

    return 1;
}


/* Gets a state (up/down) of the interface.
   Returns 1 on success, 0 on error */
static int get_iface_state(int* state)
{
    int skfd;
    struct ifreq ifr;

    if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
        error_errno("socket failed\n");
        return 0;
    }
    memcpy(ifr.ifr_name, iface, IFNAMSIZ);
    if (ioctl (skfd, SIOCGIFFLAGS, (char *) &ifr) < 0)
    {
        error_errno("ioctl SIOCGIFFLAGS failed\n");
        return 0;
    }
    if (ifr.ifr_flags & IFF_UP)
        *state = 1;
    else
        *state = 0;
    close(skfd);
    return 1; /* ok */
}


/* Sets a new state (up/down) of the interface.
   Returns 1 on success, 0 on error. */
static int set_iface_state(const int new_state)
{
    int skfd;
    struct ifreq ifr;

    if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
        error_errno("socket failed\n");
        return 0;
    }
    memcpy(ifr.ifr_name, iface, IFNAMSIZ);
    if (ioctl(skfd, SIOCGIFFLAGS, (char *) &ifr) < 0)
    {
        error_errno("ioctl SIOCGIFFLAGS failed\n");
        close(skfd);
        return 0;
    }

    if (new_state > 0)
        ifr.ifr_flags |= IFF_UP;
    else
        ifr.ifr_flags &= ~IFF_UP;

    if (ioctl(skfd, SIOCSIFFLAGS, (char *) &ifr) < 0)
    {
        error_errno("ioctl SIOCSIFFLAGS failed\n");
        close(skfd);
        return 0;
    }
    close(skfd);
    return 1; /* ok */
}

/* Checks whether two given channels are neighbouring channels
   on a supported channels range taking into account the
   IEEE mode. Usually in B mode nearest channel is 1 channel away,
   and in A mode it is not so trivial.
   Returns 1 if neighbours, 0 otherwise. */
static int are_neighbours(const int ch1, const int ch2)
{
    int neigh_dist; /* distance between 2 neighbours */

    if (ch1 == 0 || ch2 == 0)
    	return 0;

    /* Consider the distance to use according to IEEE mode */
    switch (ieee_mode)
    {
    case IEEE_MODE_A:
    	neigh_dist = CHANNELS_DISTANCE_A;
    	break;
    case IEEE_MODE_B:
    case IEEE_MODE_G:
    default:
    	neigh_dist = CHANNELS_DISTANCE_BG;
    	break;
    }

    if (abs(ch2 - ch1) <= neigh_dist)
    	return 1; /* are neighbours */

    return 0;
}

/* Calculates neighbouring channels qualities */
static void calc_neighbours_quality(void)
{
    int i, j;

    /* Process all the channels */
    for (i = 0;  i < QUAL_DATA_SIZE;  i++)
    {
        int ch_i = qual_data[i].channel;

        if (ch_i == 0)
            continue; /* no channel */

        for (j = 0;  j < QUAL_DATA_SIZE;  j++)
        {
            int ch_j = qual_data[j].channel;

            if (ch_j == 0)
                continue; /* no channel */

            if (ch_i == ch_j)
                continue; /* the same channel */

            if (are_neighbours(ch_i, ch_j))
            {
                qual_data[i].neighbours_quality += qual_data[j].my_quality;
            }
        }
    }
}

/* Sorts channels in qual_data array based on my_quality value.
 * Channel with the lowest quality is first. */
static void sort_my_quality(void)
{
    int is_sorted = 0;
    channel_quality tmp;

    while (!is_sorted)
    {
        int i;

        is_sorted = 1;
        for (i = 0;  i < QUAL_DATA_SIZE - 1;  i++)
        {
            if (qual_data[i].my_quality > qual_data[i + 1].my_quality)
            {
            	/* Swap the channels in array */
                tmp = qual_data[i];
                qual_data[i] = qual_data[i + 1];
                qual_data[i + 1] = tmp;

                /* Continue sorting */
                is_sorted = 0;
            }
        }
    }
}

/* Dumps the qual_data array to debug output */
static void print_qualities(void)
{
    int i;

    debug("Scanned radio channel qualities:\n");
    debug("[Ch]\tQual\tCount\tNeigh\n");
    for(i = 0;  i < QUAL_DATA_SIZE;  i++)
    {
        debug("[%u]\t%lu\t%u\t%lu\n",
              qual_data[i].channel,
              qual_data[i].my_quality,
              qual_data[i].ap_count,
              qual_data[i].neighbours_quality);
    }
}

/* Finds a channel is qual_data and updates its characteristics */
static void update_my_quality(const int ch, const int level)
{
    int i;

    for (i = 0; i < QUAL_DATA_SIZE; i++)
    {
        if (qual_data[i].channel == ch)
        {
            qual_data[i].my_quality += level;
            qual_data[i].ap_count++;
        }
    }
}

/* Processes the scan results: extracts channel no, level.
   Returns 1 if results received, 0 means no results. */
static int process_scan_results(const wireless_scan_head* context)
{
    wireless_scan* res; /* scanning result, info about an AP */
    int res_count = 0;  /* correct results count */

    res = context->result;
    while (res)
    {
        int ch = 0;            /* current channel */
        int level = 0;         /* current level */

        /* Remember result address for cleanup */
        wireless_scan* prev = res;

        if (res->has_stats)
            level = res->stats.qual.level;

        if (res->b.has_freq)
            ch = iw_freq_to_channel(res->b.freq, &range);

        if (level > 0 && ch > 0)
        {
            /* Update the qual_data array */
            update_my_quality(ch, level);

            debug("essid:%-20s\tfreq:%f\tchannel:%d\tlevel:%d\n",
                  res->b.has_essid ? res->b.essid : "",
                  res->b.freq, ch, level);

            res_count++;
        }

        /* Clean up after usage and proceed to a next linked list object */
        res = res->next;
        free(prev);
    }

    if (res_count)
        return 1; /* at least 1 correct result got */

    return 0; /* no results */
}

/* Triggers the channel scanning.
   Returns 1 on success, 0 on error. */
static int scan_channels(wireless_scan_head* context)
{
    int skfd = 0; /* socket fd */

    /* Open socket */
    skfd = iw_sockets_open();
    if (skfd < 0)
    {
        error_errno("iw_sockets_open failed\n");
        return 0;
    }

    /* Perform blocking scan */
    if (iw_scan(skfd, iface, range.we_version_compiled, context) < 0)
    {
        error_errno("iw_scan failed\n");
        iw_sockets_close(skfd);
        return 0;
    }

    iw_sockets_close(skfd);

    return 1; /* ok */
}

/* Collects card info: range struct, channels and ieee_mode.
   Returns 1 on success, 0 on error. */
static int collect_card_info(void)
{
    int skfd = 0; /* socket fd */
    struct iwreq  wrq;

    /* Open socket */
    skfd = iw_sockets_open();
    if (skfd < 0)
    {
        error_errno("iw_sockets_open failed\n");
        return 0;
    }

    if (iw_get_range_info(skfd, iface, &range) < 0)
    {
        error_errno("iw_get_range_info failed\n");
        iw_sockets_close(skfd);
        return 0;
    }

    /* Fill qual_data channels */
    if (range.num_frequency > 0)
    {
        int i;

        debug("%d channels in total\n", range.num_frequency);
        /* card_ch_count = range.num_frequency; */
        for (i = 0; i < range.num_frequency; i++)
        {
             qual_data[i].channel = range.freq[i].i;
             debug("No. %d Channel %.2d\n", i, range.freq[i].i);
        }
    }

    /* Determine the IEEE mode */
    if (iw_get_ext(skfd, iface, SIOCGIWFREQ, &wrq) >= 0)
    {
        double freq;
        int ch;

        freq = iw_freq2float(&(wrq.u.freq));
        ch = iw_freq_to_channel(freq, &range);

        /* Quite a dirty way to determine mode. Meanhwile I don't know
           how to distinguish between B and G modes but this is not
           an issue at the moment since B and G use the same channels.
           Let's say that if current channel > 14 than its A mode,
           otherwise its B. Please suggest a propper way to determine
           the mode. Private ioctl doesnt fit since it is not universal. */
        if (ch > 14)
            ieee_mode = IEEE_MODE_A;
        else
            ieee_mode = IEEE_MODE_B;

        debug("Current channel %d, IEEE mode: %s\n",
              ch, (ieee_mode == IEEE_MODE_A) ? "A" : "B or G");
    }

    iw_sockets_close(skfd);

    return 1; /* ok */
}

/*
   Selects the "best" channel to use.

   wiface   - wireless interface name;
   channels - a subset of allowed by user channels;
   count    - count of allowed channels;

   Returns: "best" channel no. on success,
            0 on error or no usable channel found.
*/
int get_best_channel(const char* wiface, const int count, const int* channels)
{
    int idx = 1, best = 0;
    long prev_mode;
    int prev_state;
    int i, j;
    int rv = 0; /* return value */
    wireless_scan_head context;

    /* Check input params */
    if (!wiface || !*wiface || count <= 0 || !channels)
    {
        debug("Invalid params supplied.\n");
        return 0;
    }

    /* Initialize globals */
    init();
    strncpy(iface, wiface, IFNAMSIZ);

    /* Remember previous interface state and go down. Do not check for return
       values since interface state is not deadly important. */
    get_iface_state(&prev_state);

    /* Remember previous operation mode(master/managed) */
    if (!get_mode(&prev_mode))
        goto exit;

    /* Go to managed(client) mode */
    if (!set_mode(IW_MODE_INFRA))
        goto exit;

    /* Collect all the necessary data from the driver */
    if (!collect_card_info())
    {
        error("collect_card_info failed.\n");
        goto exit;
    }

    /* Now we can proceed with scanning */
    set_iface_state(IFACE_UP);
    if (!scan_channels(&context))
    {
        error("scan_channels failed.\n");
        goto exit;
    }

    /* Process scan results: fill qual_data array */
    if (!process_scan_results(&context))
    {
        debug("Wireless scan succeeded but with no usable results.\n");

        /* Return the last of the prefered channels */
        rv = channels[count - 1];

        debug("Autochannel best channel: %d\n", rv);
        goto exit;
    }

    calc_neighbours_quality();
    print_qualities();

    /* First "disable" unused channels (set quality to max possible value) */
    for (i = 0;  i < QUAL_DATA_SIZE; ++i)
    {
        int disable = 1;

        /* Don't need to disable if channel is in a user's list */
        for (j = 0;  j < count;  ++j)
        {
            if (qual_data[i].channel == (unsigned int)channels[j])
                disable = 0;
        }

        if (disable)
        {
            qual_data[i].my_quality = ~0;
            qual_data[i].neighbours_quality = ~0;
        }
    }

    /* Now sort qualities (unused qualities will be the last) */
    sort_my_quality();

    /* And the last step - get the best channel. This one is the trickiest
       step. */

    /* Compare only the channels that have "near the best" quality */
    while (abs(qual_data[idx].my_quality - qual_data[0].my_quality) < QUALITY_DELTA)
    {
        /* It seems that two channels are similar by their
           quality levels, let's check for AP count. Remember
           that AP count on a channel is higher priority than
           quality level. */
        if (qual_data[idx].ap_count < qual_data[best].ap_count)
        {
            /* And YES! A current channel has less APs than the "best".
             * Let's change the "best" channel to a current. */
            best = idx;
        }
        else if (qual_data[idx].ap_count == qual_data[best].ap_count)
        {
            /* Two channels have the same amount of APs. Let's
             * check their neighbours. The winning channel is
             * the one with less crowded neighbours. */
            if (qual_data[idx].neighbours_quality <=
                qual_data[best].neighbours_quality)
            {
                best = idx;
            }
        }

        /* Proceed to next channel */
        idx++;

        /* Maybe we are finished? */
        if (idx >= QUAL_DATA_SIZE)
            break;
    }

    rv = qual_data[best].channel;
    debug("Autochannel best channel: %d\n", qual_data[best].channel);

exit:

    /* Return interface state to the initial */
    set_iface_state(prev_state);

    /* Return operation mode to the initial */
    set_mode(prev_mode);

    return rv;
}

